"""
Module to retrieve location data based on 
- GPS Coordinates
- Address
- Landmark name
- Zip code

NOTE:

Requires a FREE API Key from https://geocode.maps.co

See set_api_tokens.exe (dt-cli-tools) for caching the token.
"""
import json
import pathlib
from dataclasses import dataclass, field
from datetime import datetime
from time import sleep
from typing import Dict, Tuple

import requests
from loguru import logger as LOGGER
from timezonefinder import TimezoneFinder

import dt_tools.logger.logging_helper as lh
from dt_tools.misc.api_helper import ApiTokenHelper as api_helper

# ============================================================================================
class _GeoLoc_Control:
    API_KEY = api_helper.get_api_token('geocode.maps.co')
    API_ENABLED = API_KEY is not None
    BASE_URL = 'https://geocode.maps.co'
    ADDRESS_URI = 'search'
    LAT_LON_URI = 'reverse'
    GEOLOC_CACHE_FILENM = pathlib.Path('~').expanduser().absolute() / ".IpHelper" / 'geoloc_cache.json'

# ============================================================================================
@dataclass
class LocationCache:
    cache: Dict[str, dict] = field(default_factory=dict)
    valid_cache: bool = True

    def __post_init__(self):
        if not _GeoLoc_Control.API_ENABLED:
            LOGGER.trace('GeoLoc API key not defined, cache lookup only.')

        self.load()
        if self.valid_cache:
            LOGGER.trace(f'Location cache loaded with {len(self.cache)} entries.')
        else:
            LOGGER.trace('Cache does not exist and could not be created.')

    def __del__(self):
        if self.valid_cache:
            self.save()
            LOGGER.debug(f'Location cache saved with {len(self.cache)} entries to {_GeoLoc_Control.GEOLOC_CACHE_FILENM}')

    def save(self):
        if self.valid_cache:
            cache_file = _GeoLoc_Control.GEOLOC_CACHE_FILENM
            cache_file.write_text(json.dumps(self.cache,indent=2), encoding='UTF-8')

    def load(self):
        cache_file = _GeoLoc_Control.GEOLOC_CACHE_FILENM
        if cache_file.exists():
            self.cache = json.loads(cache_file.read_text(encoding='UTF-8'))
        elif cache_file.parent.exists():
            self.cache = {}
        else:
            cache_file.parent.mkdir(parents=True)

    @lh.logger_wraps()
    def clear(self):
        cache_file = pathlib.Path(_GeoLoc_Control.GEOLOC_CACHE_FILENM)
        if cache_file.exists():
            LOGGER.warning(f'{len(self.cache)} GeoLoc cache entries cleared.')
            cache_file.unlink(missing_ok=True)
            self.cache = {}

    def exists(self, key) -> bool:
        return key in self.cache.keys()
    
    def get(self, key) -> dict:
        return self.cache.get(key, None)
    
    def add(self, key, data):
        if self.exists(key):
            LOGGER.error(f'{key} - ALREADY EXISTS IN CACHE')
        self.cache[key] = data

LOCATION_CACHE: LocationCache = LocationCache()

# ============================================================================================
class GeoLocation:
    """
    Class to represent GPS Location, with interfaces to populate via 
    - GPS Location (latitude, longitude)
    - Address (House, Street, City, State, Zip)
    - Address (formattet string)
    - Landmark (For example- Statue of Liberty or Mount Rushmore)
    - Zip 

    """
    lat: float = None
    lon: float = None
    display_name: str = None
    house: int = None
    street: str = None
    city: str = None
    county: str = None
    state: str = None
    zip: int = None
    ip: str = None
    tz_name: str = None

    _tf: TimezoneFinder = TimezoneFinder()
    _json_payload: dict = None
    _last_call: datetime = datetime.now()

    @property
    def address(self) -> str:
        if self.city is None and self.state is None:
            return ''
        address = ''
        if self.house:
            address += f'{self.house} '
        if self.street:
            address += f'{self.street}, '
        
        if self.city:
            address += f"{self.city}, "
        elif self.county:
            address += f'{self.county.replace("County","")}, '
        
        if self.state:
            address += f'{self.state}, '
        if self.zip:
            address += f'{self.zip}, '

        if self.country:
            address += f'{self.country.upper()}'

        return address
        
    @property
    def lat_lon(self) -> str:
        """Return latitude longitude as a comma seperated string"""
        if self.lat is None or self.lon is None:
            return ''
        return f'{float(self.lat):.7f},{float(self.lon):.7f}'

    @lh.logger_wraps()
    def _clear_location_data(self):
        self.lat = None
        self.lon = None
        self.display_name = None
        self.house = None
        self.street = None
        self.city = None
        self.county = None
        self.state = None
        self.zip = None
        self.country = None
        self.ip = None
        self.tz_name = None
        self._json_payload = None
        LOGGER.trace('Location data cleared.')

    @lh.logger_wraps()
    def _load_location_data_from_cache(self, key: str) -> bool:
        if LOCATION_CACHE.exists(key):        
            loc_dict = LOCATION_CACHE.get(key)
            self.lat = loc_dict.get('lat')
            self.lon = loc_dict.get('lon')
            self.display_name = loc_dict.get('display_name')
            self.house = loc_dict.get('house')
            self.street = loc_dict.get('street')
            self.city = loc_dict.get('city')
            self.county = loc_dict.get('county')
            self.state = loc_dict.get('state')
            self.zip = loc_dict.get('zip')
            self.country = loc_dict.get('country')
            self.ip = loc_dict.get('ip')
            self.tz_name = loc_dict.get('tz_name')
            LOGGER.debug(f'({self.lat_lon}) retrieved from cache')
            return True
        
        return False
    
    @lh.logger_wraps()
    def get_location_via_lat_lon(self, lat: float, lon: float) -> bool:
        """Retrieve address location based on lat/lon coordinates"""
        self._clear_location_data()
        self.lat = lat
        self.lon = lon
        loc_dict = None
        if self._load_location_data_from_cache(self.lat_lon):
            LOGGER.debug('-> Loaded from cache.')
        elif not _GeoLoc_Control.API_ENABLED:
            return False
        else:
            url = f"{_GeoLoc_Control.BASE_URL}/{_GeoLoc_Control.LAT_LON_URI}?api_key={_GeoLoc_Control.API_KEY}&lat={self.lat}&lon={self.lon}"
            LOGGER.trace(f'GEOLOC url: {url}')
            loc_dict = self._api_call(url)

        if loc_dict:
            addr_dict: dict = loc_dict.get('address', {})
            self.display_name = loc_dict['display_name']
            self.house = addr_dict.get('house_number', None)
            token = addr_dict.get('road', None)
            if token is None:
                token = addr_dict.get('street', None)
            self.street = token
            token = addr_dict.get('city', None)
            if token is None:
                token = addr_dict.get('own', None)
            if token is None:
                token = addr_dict.get('hamlet', None)
            self.city = token
            self.county = addr_dict.get('county', None)
            self.state = addr_dict.get('state', None)
            self.country = addr_dict.get('country_code', None)
            self.zip = addr_dict.get('postalcode', None)
            # amenity: xxxx,  residential: cimarrone, village: Spring Lake, municipality: Spring Lake Township, neighbourhood: south slope, suburb: brooklyn, city: new york, state: new york
            return True
        
        return False
    
    @lh.logger_wraps()
    def get_location_via_address_string(self, address: str, clear_existing: bool = True) -> bool:
        """
        Retrieve location based on street address

        Required:
            address : typically house street, city, state, zip

        Returns:
            bool: True if location identified, False if not found.
        """
        if not _GeoLoc_Control.API_ENABLED:
            return False

        if clear_existing:
            self._clear_location_data()
        url = f"{_GeoLoc_Control.BASE_URL}/{_GeoLoc_Control.ADDRESS_URI}?api_key={_GeoLoc_Control.API_KEY}&q={address}"
        LOGGER.trace(f'GEOLOC url: {url}')
        loc_dict = self._api_call(url)
        self._populate_via_payload(loc_dict)
        return loc_dict is not None # Found
        
    def get_location_via_landmark(self, landmark: str) -> bool:
        """
        Retrieve location based on landmark name.

        Required:
            landmark (str): Landmark name (ie. Statue of Liberty)

        Returns:
            bool: True if location identified, False if not found.
        """
        if not _GeoLoc_Control.API_ENABLED:
            return False
        
        return self.get_location_via_address_string(address=landmark)
        
    @lh.logger_wraps()
    def get_location_via_address(self, city: str, state: str, house: int=None, street: str=None, zip: int=None) -> bool:
        """
        Retrieve location based on street address
        
        Required:
            city, state

        Optional:
            house, street, zip

        Returns:
            bool: True if location identified, False if not found.
        """
        self._clear_location_data()
        if not _GeoLoc_Control.API_ENABLED:
            return False
        
        # Load fields so self.address populates
        self.house = house
        self.street = street
        self.city = city
        self.state = state
        self.zip = zip
        return self.get_location_via_address_string(self.address, clear_existing=False)
    
    
    @lh.logger_wraps()
    def get_location_via_zip(self, zip: str, country_cd: str = None) -> bool:
        """Retrieve location based on zip code"""
        self._clear_location_data()
        self.zip = zip
        if not _GeoLoc_Control.API_ENABLED:
            return False

        url = f"{_GeoLoc_Control.BASE_URL}/{_GeoLoc_Control.ADDRESS_URI}?api_key={_GeoLoc_Control.API_KEY}&postalcode={self.zip}"
        if country_cd is not None:
            url = f'{url}&country={country_cd}'
        loc_dict = self._api_call(url)
        if loc_dict:
            self._populate_via_payload(loc_dict)
            return True
        
        return False
    
    @lh.logger_wraps()
    def get_location_via_ip(self) -> bool:
        """Retrieve location info based on IP address"""
        self._clear_location_data()
        # Retrieve public IP address
        resp = requests.get('http://ip-api.com/json/')
        resp_json = resp.json()
        if resp_json.get('status') == "success":
            ip = resp_json.get('query')
            self.lat = resp_json.get('lat')
            self.lon = resp_json.get('lon')
            self.city = resp_json.get('city')
            self.country = resp_json.get('countryCode')
            self.state = resp_json.get('region')
            self.zip = resp_json.get('zip')
            self.ip = ip
            self.tz_name = resp_json.get('timezone')
            self._json_payload = resp_json
            LOGGER.debug(f'External IP identified as: {ip}')
            return True
        
        self.lat = 0.0
        self.lon = 0.0
        LOGGER.error('Unable to determine ip for location identification')
        return False

    # ---------------------------------------------------------------------------------
    @lh.logger_wraps()
    def _api_call(self, url) -> Dict:
        while (datetime.now() - self._last_call).total_seconds() < 1.1:
            LOGGER.trace('throttle.')
            sleep(.1)

        LOGGER.trace(f'GEOLOC url: {url}')
        loc_dict: dict = None
        throttle = True
        while throttle:
            try:
                resp = requests.get(url)
                if resp.status_code == 200:
                    json_data = resp.json()                
                    LOGGER.trace(resp.json())
                    throttle = False
                    if isinstance(json_data, dict):
                        loc_dict = json_data
                    elif isinstance(json_data, list):
                        if len(json_data) > 0:
                            loc_dict = json_data[0]
                        else:
                            LOGGER.warning(f'GEOLOC not found for {url}')
                    else:
                        LOGGER.error(f'Unknown response: {url} - {json_data}')
                elif resp.status_code == 429:
                    LOGGER.debug('GEOLOC throttle...')
                    sleep(1.1)
                else:
                    throttle = False
                    LOGGER.warning(f'URL: {url}  RC: {resp.status_code} - {resp.text}')            

            except Exception as ex:
                LOGGER.exception(f'Unable to get geoloc: {url} - {repr(ex)}')

        if loc_dict:
            self._json_payload = loc_dict
            self.lat = float(loc_dict['lat'])
            self.lon = float(loc_dict['lon'])
            self.tz_name = self._tf.timezone_at(lat=self.lat, lng=self.lon)
            loc_dict['tz_name'] = self.tz_name
            if not LOCATION_CACHE.exists(self.lat_lon):
                LOCATION_CACHE.add(self.lat_lon, loc_dict)
                LOGGER.debug(f'({self.lat_lon}) added to cache, {len(LOCATION_CACHE.cache)} total entries.')
        
        return loc_dict

    @lh.logger_wraps()
    def _populate_via_payload(self, payload: dict):
        if payload is None:
            return

        self.display_name = payload.get('display_name') if self.display_name is None else self.display_name
        if self.display_name is not None:
            tokens = self.display_name.split(', ')
            tokens.reverse()
            self.country = tokens[0] if self.country is None else self.country
            if tokens[1].isnumeric():
                LOGGER.debug(f' display_name: {self.display_name}')
                LOGGER.debug(f' tokens      : {tokens}')
                self.zip = tokens[1] if self.zip is None else self.zip
                self.state = tokens[2] if self.state is None else self.state
                self.county = tokens[3] if self.county is None else self.county
                if len(tokens) > 4:
                    self.city = tokens[4] if self.city is None else self.city

            self.house = None
            self.street = None
            LOGGER.debug(self.to_string())


    def GPS_dms_to_dd(lat_dms:Tuple[float, float, float], lat_ref: str, lon_dms: Tuple[float, float, float], lon_ref: str) -> Tuple[float, float]:
        """
        Return Lat/Lon decimal degrees (dd) from Lat/Long Degree, Minute Seconds (dms) coordinates.

        Args:
            lat_dms: (degree, minute, sec)
            lat_ref: N or S
            lon_dmsL (degree, minute, sec)
            lon_ref: E or W

        Returns:
            lat_dd: decimal degree - latitude 
            lon_dd: decimal degree - longitude
            (0.0, 0.0) on ERROR
        """
        try:
            #          Degree   + (min/60)        + (seconds/3600)
            lat_dd = float(lat_dms[0]) + (float(lat_dms[1])/60) + (float(lat_dms[2])/3600)
            lon_dd = float(lon_dms[0]) + (float(lon_dms[1])/60) + (float(lon_dms[2])/3600)
            if lat_ref.lower() == "s":
                lat_dd *= -1
            if lon_ref.lower() == "w":
                lon_dd *= -1
        except Exception as ex:
            LOGGER.error(f'Lat/Lon: {repr(ex)}')
            lat_dd = 0.0
            lon_dd = 0.0

        lat_dd = float(f'{lat_dd:.4f}')
        lon_dd = float(f'{lon_dd:.4f}')
        
        return (lat_dd, lon_dd)

    def to_string(self) -> str:
        output = ''
        for attr in dir(self):
            value = getattr(self, attr)
            if value is not None and not callable(value) and not attr.startswith('_'):
                output += f'{attr:15} : {getattr(self, attr)}\n'
        
        return output

def _print_object(obj):
    for line in obj.to_string().splitlines():
        LOGGER.info(f'  {line}')
    payload = getattr(obj,'_json_payload')
    if payload is not None:
        LOGGER.info('  Payload-')
        for k, v in payload.items():
            LOGGER.info(f'    {k:15} {v}')

if __name__ == "__main__":
    import sys

    lh.configure_logger(log_level='INFO', brightness=False)
    if '-c' in sys.argv:
        LOCATION_CACHE.clear()
        
    helper = GeoLocation()

    print('')
    address = '4833 Nahane Way, Saint Johns, FL'
    LOGGER.warning(f'Address: {address}')
    helper.get_location_via_address_string(address)
    LOGGER.info('Returns:')
    _print_object(helper)

    print('')
    lat = 30.069157
    lon = -81.551387
    LOGGER.warning(f'Location for {lat},{lon}')
    helper.get_location_via_lat_lon(lat=lat, lon=lon)
    LOGGER.info('Returns:')
    _print_object(helper)

    print('')
    zip = 32259
    LOGGER.warning(f'Zip: {zip}')
    helper.get_location_via_zip(zip, "US")
    LOGGER.info('Returns:')
    _print_object(helper)

    print('')
    address = "Matterhorn"
    LOGGER.warning(f'Address: {address}')
    helper.get_location_via_address_string(address)
    LOGGER.info('Returns:')
    _print_object(helper)



"""
weather.py - current, forecasts, alerts

This module leverages 3rd party service (weatherapi.com) to retrieve detailed weather
information based on GeloLocation (latitude, longitude).

Location information can be retrieved based on:
- GeoLocation (lat/lon)
- US address (string or components (street, city, state, zip))
- Internet IP address

"""
import datetime
import json
import pathlib
from dataclasses import dataclass
from datetime import datetime as dt

import requests
from loguru import logger as LOGGER

import dt_tools.net.net_helper as nh
from dt_tools.geoloc.census_geoloc import Census_GeoLocation
from dt_tools.misc.api_helper import ApiTokenHelper
from dt_tools.geoloc.sun import Sun
from dt_tools.weather.common import AQI_DESC, WeatherLocation, WeatherSymbols


class CURRENT_WEATHER_SETTINGS:
    API_KEY = ApiTokenHelper.get_api_token(ApiTokenHelper.API_WEATHER_INFO)
    API_AVAILABLE = False if API_KEY is None else True
    BASE_URL = "http://api.weatherapi.com/v1" # 1 million calls per month
    CURRENT_URI = "current.json"
    FORECAST_URI = "forecast.json"
    SEARCH_URI = "search.json"

@dataclass
class CurrentConditions():
    """
    Weather condition class

    Raises:
        ValueError: _description_

    Returns:
        _type_: _description_
    """
    location: WeatherLocation = None
    condition: str = None
    _condition_icon: str = None
    temp: float = None
    feels_like: float = None
    wind_direction: str = None
    wind_speed_mph: float = None
    wind_gust_mph: float = None
    humidity_pct: int = None
    cloud_cover_pct: int = None
    visibility_mi: float = None
    precipitation: float = None
    last_update: dt = None
    aqi: int = None
    aqi_text: str = None
    sunrise: datetime.datetime = None
    sunset: datetime.datetime = None
    _connect_retries: int = 0
    _disabled: bool = True

    def __post_init__(self):
        if not CURRENT_WEATHER_SETTINGS.API_AVAILABLE:
            msg = f'No API Token set for {CURRENT_WEATHER_SETTINGS.BASE_URL}.\n'
            msg += 'Use "set_api_tokens" to cache FREE API token.'
            raise ConnectionError(msg)
 
    def set_location_via_lat_lon(self, lat: float, lon: float) -> bool:
        """
        Set weather location based on Geolocation

        Args:
            lat (float): Latitude
            lon (float): Longitude

        Returns:
            bool: True if location successfully set.
        """
        if CURRENT_WEATHER_SETTINGS.API_AVAILABLE:
            self.location = WeatherLocation(lat, lon)
            return self.refresh()
    
        return False
    
    def set_location_via_census_address(self, street: str, city: str = None, state: str = None, zipcd: str = None) -> bool:
        """
        Set location based on address: street, city, state or street, zipcd

        The GeoLocation will be derived based on the address.  NOTE: only US addresses are allowed.

        Args:
            street (str): House number and street name
            city (str, optional): City. Defaults to None.
            state (str, optional): State. Defaults to None.
            zipcd (str, optional): Zip code. Defaults to None.

        Returns:
            bool: True if address is resolved and GeoLocation identified, else False
        """
        # if CURRENT_WEATHER_SETTINGS.API_AVAILABLE:
        #     geo_locs = GeoLocation.lookup_address(street=street, city=city, state=state, zipcd=zipcd)
        #     if len(geo_locs) > 0:
        #         loc = geo_locs[0]
        #         self.location = WeatherLocation(latitude=loc.latitude, longitude=loc.longitude, location_name=loc.address)
        #         return self.refresh()
    
        geo_locs = Census_GeoLocation.lookup_address(street=street, city=city, state=state, zipcd=zipcd)
        if len(geo_locs) > 0:
            loc = geo_locs[0]
            self.location = WeatherLocation(latitude=loc.latitude, longitude=loc.longitude, location_name=loc.address)
            return self.refresh()
        return False
    
    def set_location_via_address(self, address: str) -> bool:
        """
        Set location based on address string

        GeoLocation will be derived based on the address. 

        Args:
            address (str): Address string - i.e. 123 somestree, somecity, somestate, some zip

        Returns:
            bool: True if address is resolved and GeoLocation identified, else False
        """
        from dt_tools.geoloc.geoloc import GeoLocation as GeoLoc
        if CURRENT_WEATHER_SETTINGS.API_AVAILABLE:
            geo = GeoLoc()
            if geo.get_location_via_address_string(address):
                self.location = WeatherLocation(latitude=geo.lat, longitude=geo.lon, location_name=address)
                return self.refresh()
    
        return False

    def set_location_via_ip(self, ip: str = None) -> bool:
        """
        Set location based on IP address

        The IP should be resolvable on the internet (i.e. local addresses won't work)

        Args:
            ip (str, optional): IP to resolve, if None, The device external address (i.e. from
                service provider will be used). Defaults to None.

        Returns:
            bool: True if IP is resovled to GeoLocation, else False
        """
        if CURRENT_WEATHER_SETTINGS.API_AVAILABLE:
            if ip is None:
                lat, lon = nh.get_lat_lon_for_ip(ip=nh.get_wan_ip()) # self._get_lat_lon_from_ip(nh.get_wan_ip())
            else:
                lat, lon = nh.get_lat_lon_for_ip(ip=ip) # self._get_lat_lon_from_ip(ip)

            self.location = WeatherLocation(lat, lon)
            return self.refresh()
    
        return False


    @property
    def loc_name(self) -> str:
        return '' if self.location is None else self.location.location_name
    
    @loc_name.setter
    def loc_name(self, val: str):
        if self.location is not None:
            self.location.location_name = val

    @property
    def loc_region(self) -> str:
        return '' if self.location is None else self.location.location_region
    
    @loc_region.setter
    def loc_region(self, val: str):
        if self.location is not None:
            self.location.location_region = val
            
 
    @property
    def disabled(self) -> bool:
        """
        Check if location defined, API and connection is available

        Returns:
            bool: False if connection is available, else False
        """
        return self._disabled
    
    @property
    def condition_icon(self) -> str:
        return self._condition_icon
    
    @condition_icon.setter
    def condition_icon(self, value):
        LOGGER.trace(f'icon: {value}')

        icon_filenm_local = value.replace('//cdn.weatherapi.com','./files/icons')
        icon_file = pathlib.Path(icon_filenm_local).absolute()
        if icon_file.exists():
            self._condition_icon = str(icon_file)
        else:
            self._condition_icon = value

    @property
    def lat_long(self) -> str:
        return '' if self.location is None else f'{self.location.latitude},{self.location.longitude}'

    def to_string(self) -> str:
        """
        String representation of the current weather.

        Returns:
            str: _description_
        """
        degree = WeatherSymbols.degree.value
        text =  f'{self.condition}\n'    
        text += f'  Temperature {self.temp}{degree} feels like {self.feels_like}{degree}\n'    
        text += f'  Humidity {self.humidity_pct}%\n'    
        text += f'  Air Quality {self.aqi_text}\n'    
        text += f'  Precipitation {self.precipitation} in.\n'    
        text += f'  Cloud Cover {self.cloud_cover_pct}%, visibility {self.visibility_mi} miles\n'    
        text += f'  Wind {self.wind_speed_mph} mph - {self.wind_direction} with gusts up to {self.wind_gust_mph} mph\n'
        if self.sunrise:
            text += f'  Sunrise at {self.sunrise.strftime("%I:%M %p")} / Sunset at {self.sunset.strftime("%I:%M %p %Z")}'
        return text
    
    def refresh(self, ignore_cache: bool = False) -> bool:
        """
        Refresh current weather

        Args:
            ignore_cache (bool, optional): _description_. Defaults to False.

        Returns:
            bool: _description_
        """
        if ignore_cache:
            return self._refresh_if_stale(elapsed_mins=0)
        return self._refresh_if_stale()
    
    def _refresh_if_stale(self, elapsed_mins: int = 15) -> bool:
        """
        Refresh weather data if stale.  Default is 15 monutes.
        """
        if self.location is None or not self.location.is_initialized():
            raise ValueError('ABORT - Weather location is NOT initialized.')
        
        elapsed = "UNKNOWN"
        if self.last_update is not None:
            elapsed = (dt.now() - self.last_update).total_seconds() / 60
            if elapsed < elapsed_mins:
                LOGGER.trace('Weather data NOT refreshed')
                return False
        try:
            # will fail if elapsed/last_update not set
            LOGGER.debug(f'- Weather being refreshed, last update {elapsed:.2f} minutes ago at {self.last_update}')            
        except Exception as ex:
            LOGGER.trace(f'no prior weather {ex}')
            LOGGER.debug('- Weather being refreshed, last update Unknown')

        target_url=f'{CURRENT_WEATHER_SETTINGS.BASE_URL}/{CURRENT_WEATHER_SETTINGS.CURRENT_URI}?key={CURRENT_WEATHER_SETTINGS.API_KEY}&q={self.lat_long}&aqi=yes'
        LOGGER.debug(f'WEATHER url: {target_url}')
        try:
            resp = requests.get(target_url)
            if resp.status_code == 200:
                LOGGER.debug(json.dumps(resp.json(), indent=2))
                self._load_current_conditions(resp.json())
                self._disabled = False
                if self.sunrise is None:
                    sun = Sun(self.location.latitude, self.location.longitude)
                    try:
                        self.sunrise = sun.get_gps_sunrise()
                    except Exception as ex:
                        LOGGER.error(f'Unable to get sunrise: {ex}')
                    try:
                        self.sunset  = sun.get_gps_sunset()
                    except Exception as ex:
                        LOGGER.error(f'Unable to get sunset: {ex}')
                return True

        except Exception as ex:
            LOGGER.warning('Unable to call weather api')
            LOGGER.warning(f'  URL   : {target_url}')
            LOGGER.exception(f'  ERROR : {repr(ex)}')
            self._connect_retries += 1
            if self._connect_retries > 3:
                print('dt_tools.weather.weather: Unable to reconnect to weather, disabled feature.', file=sys.stderr)
                self._disabled = True
            return False
                
        LOGGER.error(f'Request URL: {target_url}')
        LOGGER.error(f'Response status_code: {resp.status_code}')
        print('dt_tools.weather.weather: Unable to connect to weather, disabled feature.', file=sys.stderr)
        self._disabled = True
        return False

    def _load_current_conditions(self, blob: dict):
        LOGGER.debug('_load_current_conditions()')
        l_block: dict = blob['location']
        w_block: dict = blob['current']
        c_block: dict = w_block.get('condition',{})
        if self.location.location_name is None:
            self.loc_name       = l_block.get('name', '')
        else:
            self.loc_name       = self.location.location_name
        self.loc_region         = l_block.get('region', '')
        self.condition          = c_block.get('text','')  # w_block["condition"]["text"]
        self.condition_icon     = c_block.get('icon', '') # w_block["condition"]["icon"]
        self.temp               = float(w_block.get("temp_f", -1))
        self.feels_like         = float(w_block.get("feelslike_f", -1)) 
        self.wind_direction     = w_block.get("wind_dir", '')
        self.wind_speed_mph     = float(w_block.get("wind_mph", -1))
        self.wind_gust_mph      = float(w_block.get("gust_mph", -1))
        self.humidity_pct       = float(w_block.get("humidity", -1))
        self.cloud_cover_pct    = float(w_block.get("cloud", -1))
        self.visibility_mi      = float(w_block.get("vis_miles", -1))
        self.precipitation      = float(w_block.get("precip_in", -1))
        try:
            self.aqi            = int(w_block["air_quality"]['us-epa-index'])
        except Exception as ex:
            LOGGER.error(f'Unable to determine AQI: {repr(ex)}')
            self.aqi = -1

        self.aqi_text    = AQI_DESC[self.aqi]     
        self.last_update = dt.now()
        
    
    
if __name__ == "__main__":
    import dt_tools.logger.logging_helper as lh
    import sys

    log_lvl = 'DEBUG' if '-v' in sys.argv else 'INFO'
    
    lh.configure_logger(log_level=log_lvl, log_format=lh.DEFAULT_DEBUG_LOGFMT, brightness=False)

    weather = CurrentConditions() # Must create new due to throttle timer
    geo = Census_GeoLocation.lookup_address(street='1812 Edgewood', city="Berkley", state='MI')
    weather.set_location_via_lat_lon(geo[0].latitude, geo[0].longitude)
    LOGGER.success(f'Weather via lat/lon: {weather.lat_long} - {weather.loc_name}')
    for line in weather.to_string().splitlines():
        LOGGER.info(f'  {line}')
    LOGGER.info('')

    addresses = [
        'Nahane Way, Saint Johns, FL, 32259',
        'Duluth, MN',
        'Denver, CO',
        'Portland, OR',
        'Matterhorn',
        'Bora Bora',
        'Puerto Williams, Chile',
        'Longyearbyen'
        ]
    
    for address in addresses:
        weather = CurrentConditions() # Must create new due to throttle timer
        weather.set_location_via_address(address=address)
        LOGGER.success(f'Weather via address: {weather.loc_name}')
        for line in weather.to_string().splitlines():
            LOGGER.info(f'  {line}')
        LOGGER.info('')




import argparse
import pathlib
import threading
from dataclasses import asdict, dataclass
from time import sleep, time
from typing import Dict, List, Union

import numpy as np
import pyaudio
from loguru import logger as LOGGER

from dt_tools.os.os_helper import OSHelper


@dataclass
class SampleRate:
    LORES_Quality: int = 22050
    CD_Quality: int    = 44100
    DVD_Quality: int   = 48000
    HIRES_Quality: int = 88200

    @classmethod    
    def rate_values(cls) -> List[int]:
        return list(asdict(SampleRate()).values())
    @classmethod
    def rate_keys(cls) -> List[str]:
        return list(asdict(SampleRate()).keys())
    @classmethod
    def rate_dict(cls) -> Dict[str, int]:
        return asdict(SampleRate())
    

class SoundDefault:
    FRAME_COUNT: int    = 2048
    CHANNELS: int       = 1
    SAMPLE_RATE: int    = SampleRate.CD_Quality
    SOUND_THRESHOLD: int = 50 if OSHelper.is_windows() else 70
    TRIGGER_CNT: int    = 3


@dataclass
class SoundDetector():
    def __init__(self, microphone_id: int = -1,
                       frame_count: int = SoundDefault.FRAME_COUNT, 
                       channels: int = SoundDefault.CHANNELS, 
                       sample_rate: SampleRate = SoundDefault.SAMPLE_RATE,
                       sound_threshold: int = SoundDefault.SOUND_THRESHOLD, 
                       trigger_cnt: int = SoundDefault.TRIGGER_CNT,
                       sound_trigger_callback: callable = None,
                       silence_trigger_callback: callable = None):
        """
        Sound detector listens for sound via the microphone and signals either
        sound detected or silence (after sound stops).

        Args:
            microphone_id (int, optional):
                ID associated with the microphone device.  If -1, default microphone will be used.

            frame_count (int, optional): 
                Number of audio frames to capture when analyzing if a sound has been made.
                Basically a buffer size. Defaults to 1024.

            channels (int, optional): 
                Number of audio channels to capture. Defaults to 1.

            sample_rate (int, optional): 
                Number of frames captured per second . Defaults to 44100.

            snd_threshold (int, optional): 
                This threshold is a computed value (rms) of the data from the microphone.
                When rms exeeds this threshold for trigger_cnt cycles, the sound_trigger_callback is executed. 
                Defaults to 70.

            trigger_cnt (int, optional): 
                The number of cycle that a sound needs to be either above or below the sample_threshold
                for either silence or sound. Defaults to 3.

            sound_trigger_callback (callable, optional): 
                The routine to be called when sound has been detected. If not supplied, a stub routine will be
                called and will print a debug message to the console.  Defaults to None.

            silence_trigger_callback (callable, optional): 
                The routine to be called when silence has been detected.  If not supplied, a stub routine will be
                called and will print a debug message to the console. Defaults to None.

        """
        self._device_id: int        = microphone_id if microphone_id >= 0 else self.default_microphone_id
        self._frame_count: int      = frame_count 
        self._channels: int         = channels
        self._sample_rate: int      = sample_rate
        self._sound_threshold: int  = sound_threshold
        self._trigger_count: int    = trigger_cnt
        
        self._sound_trigger_callback: callable   = sound_trigger_callback
        self._silence_trigger_callback: callable = silence_trigger_callback

        if sound_trigger_callback is None:
            self._sound_trigger_callback = self._callback_sound_stub
        if silence_trigger_callback is None:
            self._silence_trigger_callback = self._callback_silence_stub

        self._device_name: str      = None
        self._format: int              = pyaudio.paInt16
        self._listening: bool          = False
        self._pyaudio: pyaudio.PyAudio = None
        self._stream = pyaudio._Stream = None
        self._monitor_thread: threading.Thread = None
        self._sample_list: list        = [0.0]

        self._capture: bool              = False
        self._capture_path: pathlib.Path = None
        self._capture_file: pathlib.Path = None
        self._current_audio_mean: int    = -1
        self._current_audio_rms: int     = -1
        self._start_time: float          = 0
        self._elapsed_secs: float        = 0
        self._name: str                  = None


    @property
    def default_microphone_id(self) -> int:
        pa = pyaudio.PyAudio()
        mic_id = -1
        try:
            host_info = pa.get_default_host_api_info()
            mic_id = int(host_info.get('defaultInputDevice'))
        except IOError as ioe:
            LOGGER.debug(f'Unable to identify default microphone id. {ioe}')
        
        return mic_id

    @property
    def default_microphone_name(self) -> str:
        if self._name is None:
            pa = pyaudio.PyAudio() 
            self._name = pa.get_default_input_device_info().get('name', 'Unknown')
        return self._name

    @property
    def microphone_id(self) -> int:
        return self._device_id
    @property
    def microphone_name(self) -> str:
        if self._device_name is None:
            try:
                dev_info = pyaudio.PyAudio().get_device_info_by_index(self.microphone_id)
                self._device_name = dev_info.get('name', 'Unknown')
            except Exception as ex:
                self._device_name = f'Unknown - {ex}'
    
        return self._device_name
    
    @property
    def default_host_api_info(self) -> Union[dict, None]:
        try:
            device = pyaudio.PyAudio().get_default_host_api_info()
        except IOError as ioe:
            LOGGER.warning(f'Audio Host Info: {ioe}')
            device = None
        return device

    @property
    def default_input_device_info(self) -> Union[dict, None]:
        try:
            device = pyaudio.PyAudio().get_default_input_device_info()
        except IOError as ioe:
            LOGGER.warning(f'Audio Input Device: {ioe}')
            device = None
        return device
    
    # @property
    # def output_device_output(self) -> Union[dict, None]:
    #     try:
    #         device = pyaudio.PyAudio().get_default_output_device_info()
    #     except IOError as ioe:
    #         LOGGER.trace(f'Audio Output Device: {ioe}')
    #         device = None
    #     return device

    @property
    def current_audio_mean(self) -> int:
        return self._current_audio_mean

    @property
    def current_audio_rms(self) -> int:
        return self._current_audio_rms

    @property
    def is_listening(self) -> bool:
        return self._listening
    

    @property
    def capture_path(self) -> Union[str, None]:
        return self._capture_path
    @capture_path.setter
    def capture_path(self, val: str):
        pth = pathlib.Path(val)
        if pth.is_dir:
            self._capture_path = pth
        else:
            LOGGER.warning(f'Unable to set capture path to {pth}')


    @property
    def capture_data(self) -> bool:
        return self._capture    
    @capture_data.setter
    def capture_data(self, enable_capture: bool):
        if self._capture:
            if enable_capture:
                LOGGER.warning('Already in capture mode')
            else:
            # Turn off
                LOGGER.info('Capture disabled.')
                self._capture = False
        else:
            if enable_capture:
                tgt_path = '.' if self.capture_path is None else self.capture_path
                c_filename = OSHelper.get_temp_filename(prefix='audioCapture',
                                                        dotted_suffix='.csv', 
                                                        target_dir=tgt_path,
                                                        keep=True)
                self._capture_file = pathlib.Path(c_filename)
                rng = [str(x) for x in list(range(self._trigger_count))]
                sample_headers = f"Sample{', Sample'.join(rng)}"                
                capture_line = f'sound_detected,threshold,np_mean,rms,{sample_headers}\n'
                with self._capture_file.open('a') as c_file:
                    c_file.write(capture_line)
                self._capture = True
                LOGGER.info(f'Capture enabled: {self._capture_file}')
            else:
                LOGGER.warning('Capture not enabled, cannot disable.')

    @property
    def elapsed_monitoring_seconds(self) -> float:
        """Number of seconds monitoring has been enabled"""
        if self._start_time > 0:
            return float(f"{(time() - self._start_time):7.2}")
        
        return self._elapsed_secs

    #-- Operate functions (Start/Stop) ------------------------------------------------------
    def start(self) -> bool:
        """
        Open the audio stream and begin listening.

        Returns:
            bool: True if listening thread triggered, else False
        """
        if self.is_listening:
            LOGGER.warning('Already listening, Killing prior instance.')
            self.stop()
        self._pyaudio = pyaudio.PyAudio()
        self._stream = self._pyaudio.open(format=self._format,
                              channels=self._channels,
                              rate=self._sample_rate,
                              input=True,
                              # input_device_index=input_device,
                              frames_per_buffer=self._frame_count)
        self._listening = True
        self._monitor_thread = threading.Thread(target=self._monitor, name='snd_monitor')
        self._monitor_thread.start()
        return True

    def stop(self) -> bool:
        """
        Close the audio stream and stop listening.

        Returns:
            bool: True if thread is stopped, False if there was an error.
        """
        if not self.is_listening:
            LOGGER.warning('Not listening.')
            return True

        if self.capture_data:
            self.capture_data = False

        self._listening = False
        try:
            self._monitor_thread.join()
            self._stream.close()
            self._pyaudio.terminate()

        except Exception as ex:
            LOGGER.error(f'stop(): {ex}')

        self._pyaudio = None
        
        return True

    #-- Set callback functions ------------------------------------------------------------
    def set_sound_callback(self, func: callable):
        """
        Set the routine to be called when a sound occurs.

        Args:
            func (callable): the function name
        """
        self._sound_trigger_callback = func
    
    def set_silence_callback(self, func: callable):
        """
        Set the routine to be called when silence occurs.

        Args:
            func (callable): the function name
        """
        self._silence_trigger_callback = func


    #== Private Functions ================================================================
    def _callback_sound_stub(self):
        LOGGER.debug(f'Sound detected.  {self._sample_list}')

    def _callback_silence_stub(self):
        LOGGER.debug(f'Silence.         {self._sample_list}')

    def _output_settings(self):
        LOGGER.debug('Sound monitoring starting.')
        LOGGER.debug(f'- Microphone ID : {self._device_id}')
        LOGGER.debug(f'           Name : {self._name}')
        LOGGER.debug(f'- Channels      : {self._channels}')
        LOGGER.debug(f'- Format        : {self._format}')
        LOGGER.debug(f'- Frame count   : {self._frame_count}')
        LOGGER.debug(f'- Sample rate   : {self._sample_rate}')
        LOGGER.debug(f'- Trigger cnt   : {self._trigger_count}')
        LOGGER.debug(f'- Snd Threshold : {self._sound_threshold}')
        LOGGER.debug(f'- Sound CB      : {self._sound_trigger_callback.__name__}')
        LOGGER.debug(f'- Silence CB    : {self._silence_trigger_callback.__name__}')

    def _monitor(self):
        self._output_settings()
        LOGGER.debug('Listening...')
        was_silent: bool = True
        sound_cnt: int  = -1
        silent_cnt: int = -1
        self._start_time = time()
        try:
            while self.is_listening:
                raw_data = self._get_audio_stream_data()
                # Convert buffer to numpy array
                np_data  = np.frombuffer(raw_data, dtype=np.short)[-self._frame_count:]
                audio_data_array = np.frombuffer(np_data, dtype=np.int16)
                sound_detected = self._is_sound_detected(audio_data_array, self._sound_threshold)
                if sound_detected:
                    silent_cnt = 0
                    if was_silent:
                        sound_cnt += 1
                        if sound_cnt >= self._trigger_count:
                            if self._sound_trigger_callback is not None:
                                self._sound_trigger_callback()
                            was_silent = False
                            sound_cnt = 0
                else: # Currently identified as silent
                    sound_cnt = 0
                    if not was_silent:
                        silent_cnt += 1
                        if silent_cnt >= self._trigger_count:
                            if self._silence_trigger_callback is not None:
                                self._silence_trigger_callback()
                            was_silent = True
                            silent_cnt = 0
        
        except Exception as ex:
            LOGGER.exception(f'Uh oh - {ex}')
            self.stop()
            
        self._elapsed_secs = float(f"{(time() - self._start_time):7.2}")
        self._start_time = 0

    def _get_audio_stream_data(self) -> bytes:
        # Loop while buffer fills
        while self._stream.get_read_available() < self._frame_count: 
            sleep(0.01)
        
        # Read buffer size number of frames
        num_frames = self._stream.get_read_available()
        frames = self._stream.read(num_frames, exception_on_overflow=False)
        
        return frames
    
    def _is_sound_detected(self, audio_data: np.ndarray, threshold) -> bool:
        np_mean = np.mean(audio_data**2)
        rms = np.sqrt(np_mean) if np_mean >= 0 else np.float64(0.0)
        sound_detected = True if rms > threshold else False

        LOGGER.trace(f'          [{rms:4.2f}] {np_mean:7.2f} {sound_detected}')
        self._current_audio_mean = np_mean
        self._current_audio_rms  = rms

        self._sample_list.append(float(f'{self._current_audio_rms:.4f}'))
        if len(self._sample_list) > self._trigger_count:
            self._sample_list = self._sample_list[-self._trigger_count:]
        if self.capture_data:
            samples = [f'{x:.4f}' for x in self._sample_list]
            capture_line = f'{sound_detected}, {threshold}, {np_mean:.4f}, {rms:.4f}, {", ".join(samples)}\n'
            with self._capture_file.open("a") as f:
                f.write(capture_line)

        return sound_detected


def _output_audio_device_report():
    pa = pyaudio.PyAudio()
    default_host_api = pa.get_default_host_api_info().get('index')
    for h_idx in range(pa.get_host_api_count()):
        api_info = pa.get_host_api_info_by_index(h_idx)
        num_devices = api_info.get('deviceCount')
        dflt_i_idx = api_info.get('defaultInputDevice')
        dflt_o_idx = api_info.get('defaultOutputDevice')
        api_name   = api_info.get('name')
        LOGGER.info('='*93)
        if h_idx == default_host_api:
            LOGGER.success(f'Host API [{h_idx:1}] - {api_name} {" [DEFAULT]" if h_idx == default_host_api else ""}')
        else:
            LOGGER.info(f'Host API [{h_idx:1}] - {api_name} {" [DEFAULT]" if h_idx == default_host_api else ""}')
        LOGGER.info(f'devices: {num_devices:2}   default input device: {dflt_i_idx:2}   default output device: {dflt_o_idx:2}')
        LOGGER.info('-'*93)
        LOGGER.info('h / d  idx Name                           ic oc  li lat  lo lat  hi lat  ho lat  Sample Rate')
        LOGGER.info('------ --- ------------------------------ -- --  ------- ------- ------- ------- -----------')
        for d_idx in range(api_info.get('deviceCount')):
            device = pa.get_device_info_by_host_api_device_index(h_idx, d_idx)
            if device.get('index') in [dflt_i_idx, dflt_o_idx]:
                log_level = "SUCCESS"
            else:
                log_level = "INFO"
            dev_idx     = device.get('index')
            dev_name    = device.get('name')
            i_channels  = device.get('maxInputChannels')
            o_channels  = device.get('maxOutputChannels')
            li_latency  = device.get('defaultLowInputLatency')
            lo_latency  = device.get('defaultLowOutputLatency')
            hi_latency  = device.get('defaultHighInputLatency')
            ho_latency  = device.get('defaultHighOutputLatency')
            sample_rate = device.get('defaultSampleRate')
            LOGGER.log(log_level,f'[{h_idx:1},{d_idx:2}] {dev_idx:3} {dev_name[:30]:30} {i_channels:2} {o_channels:2}  {li_latency:7.5f} {lo_latency:7.5f} {hi_latency:7.5f} {ho_latency:7.5f} {sample_rate:12.0f}')
    LOGGER.info('')
    LOGGER.info('LEGEND - ic    : Max Input Channels          oc    : Max Output Channels)')
    LOGGER.info('         li lat: Default Low Input Latency   lo lat: Default Low Output Latency')
    LOGGER.info('         hi lat: Default High Input Latency  ho lat: Default High Output Latency')
    LOGGER.info('')

__STOP_REQUESTED = False
def __stop_handler(signum, frame):
    global __STOP_REQUESTED
    __STOP_REQUESTED = True


if __name__ == '__main__':
    import dt_tools.logger.logging_helper as lh
    
    parser = argparse.ArgumentParser()
    parser.add_argument('-s', '--size', type=int, default=SoundDefault.FRAME_COUNT, 
        help    =f'Sample buffer size.  Default {SoundDefault.FRAME_COUNT} bytes')
    parser.add_argument('-t', '--threshold', type=int, default=SoundDefault.SOUND_THRESHOLD,
        help = f'Sound threshold.  Default {SoundDefault.SOUND_THRESHOLD}.')
    parser.add_argument('-c', '--count', type=int, default=SoundDefault.TRIGGER_CNT,
        help=f'How many time threshold needs to be exceeded to count as sound.  Default {SoundDefault.TRIGGER_CNT}.')
    parser.add_argument('-r', '--rate', type=SampleRate, default=SoundDefault.SAMPLE_RATE,
        help=f'Freq/number of frames captured per second.  Default {SoundDefault.SAMPLE_RATE}.')
    parser.add_argument('-v', '--verbose', action='count', default=0,
        help='Verbose logging (-v DEBUG, -vv TRACE)')
    args = parser.parse_args()

    if args.verbose > 1:
        log_level = "TRACE"
    elif args.verbose == 1:
        log_level = "DEBUG"
    else:
        log_level = "INFO"

    lh.configure_logger(log_level=log_level, log_format=lh.DEFAULT_DEBUG_LOGFMT)
    LOGGER.debug(f'Log level set to {log_level}')

    OSHelper.enable_ctrl_c_handler(__stop_handler)
    snd_monitor = SoundDetector(frame_count=args.size,
                                sample_rate=args.rate,
                                sound_threshold=args.threshold,
                                trigger_cnt=args.count)
    
    _output_audio_device_report()
    # snd_monitor.capture_path = './docs'
    snd_monitor.capture_data = True
    snd_monitor.start()
    while not __STOP_REQUESTED:
        sleep(1)

    snd_monitor.stop()
    LOGGER.info("That's all folks!")



"""
Speak a string of text or speak the contents of a text file.

Currently, these routines expect VLC to be installed.

Example::
    from dt_tools.misc.sound import Accent, Sound

    obj = Sound()
    obj.speak('This is a test')
    obj.speak('This is a test, with an australian accent.', accent=Accent.Australia)

ToDo:

    Update to be cross platform without relying on VLC

"""
import os
import pathlib
import textwrap
import threading
from enum import Enum
from time import sleep

from dt_tools.os.os_helper import OSHelper as helper
from gtts import gTTS
from loguru import logger as LOGGER


class Accent(Enum):
    """Accent codes for speaking"""
    Australia = "com.au"
    UnitedKingdom = "co.uk"
    UnitedStates = "us"
    Canada = "ca"
    India = "co.in"
    Ireland = "ie"
    SouthAfrica = "co.za"
    Nigeria = "com.ng"


class Sound(object):
    """
    Class to speak a string of text (or contents of a text file).

    This class relies on VLC being installed, it works on both Windows and Linux.

    Raises:
        FileNotFoundError: If file is not found.

    """
    _locked: bool = False
    _speak_thread_id: int = None
    _VLC: str = None

    def __new__(cls):
        # Make this class a singleton
        if not hasattr(cls, '_instance'):
            cls._instance = super(Sound, cls).__new__(cls)

        if not cls._is_VLC_installed():
            raise FileNotFoundError('VLC is required to use this module.  Unable to locate VLC module')
        return cls._instance


    # -- Public Functions -------------------------------------------------------------------------------------------
    @classmethod
    def speak(cls, in_token: str, speed: float = 1.0, accent: Accent = Accent.UnitedStates, ignore_in_progress: bool = False, wait: bool = True, delete_audio:bool = True) -> bool:
        """
        Speak the text string or contents of the file

        NOTE: Speech will block until done if already speaking.

        Args:
            in_token (str): 
                File or string of text to be spoken

            speed (float, optional): 
                Speed (cadence) of voice. Higher numbers faster cadence. Defaults to 1.0.

            accent (Accent, optional): 
                Accent of speaker. Defaults to Accent.UnitedStates.

            ignore_in_progress: (bool, optional): 
                Ignore request if speech is already in progress. Defaults to False

            wait: (bool, optional): 
                Wait for speech to finish before returning. Defaults to True.

            delete_audio (bool, optional): 
                Remove generated audio file. Defaults to True

        Returns:
            bool: True if successful else False
        """
        if cls.is_speaking() and ignore_in_progress:
            LOGGER.warning(f'Speak thread [{cls._speak_thread_id}] in process... Ignoring request.')
            return False
        
        if cls.is_speaking():
            LOGGER.debug(f'Waiting to speak... {in_token}')
            while cls.is_speaking():
                sleep(.25)

        cls._locked = True
        text = pathlib.Path(in_token).read_text() if cls._is_file(in_token) else in_token

        kwargs = {'text': text, 'speed': speed, 'accent': accent, 'delete_audio': delete_audio}
        t = threading.Thread(target=cls._speak, kwargs=kwargs)
        t.start()
        cls._speak_thread_id = t.native_id
        
        if wait:
            while cls.is_speaking():
                sleep(.25)

    @classmethod
    def is_speaking(cls) -> bool:
        """
        Is speech in progress...

        Returns:
            bool: True if speech is occuring, else False
        """
        return True if cls._speak_thread_id is not None or cls._locked else False
    
    @classmethod
    def play(cls, sound_file: str, speed: float = 1.0) -> int:
        """
        Play a sound file.

        Args:
            sound_file (str): Filename
            speed (float, optional): Speed (cadence) of voice. Defaults to 1.0.

        Returns:
            int: 0 if successful else non-zero
        """
        while cls.is_speaking():
            LOGGER.trace('waiting..')
            sleep(1)
        cls._locked = True
        result = cls._play(sound_file, speed)
        cls._locked = False
        return result
    
    # -- Private Functions -------------------------------------------------------------------------------------------
    @classmethod
    def _speak(cls, text: str, speed: float, accent: Accent, delete_audio: bool = True) -> int:
        sound_file = helper.get_temp_filename(prefix='dt-', dotted_suffix='.wav')
        LOGGER.debug(f'Speak thread {cls._speak_thread_id} started.')
        # tld top level domain for English
        # com.au (Australian), co.uk (United Kingdom), us (United States),    ca (Canada), 
        # co.in (India),       ie (Ireland),           co.za (South Africa),  com.ng (Nigeria)
        tts_obj = gTTS(text=text, lang='en', tld=accent.value, slow=False)
        LOGGER.debug(f'save {sound_file}')
        tts_obj.save(sound_file)
        
        display_text = textwrap.wrap(text=text, width=100, initial_indent='- Speak: ', subsequent_indent='         ')
        for line in display_text:
            LOGGER.trace(line)
        ret = cls._play(sound_file, speed)
        try:
            pathlib.Path(sound_file).unlink()
        except Exception as ex:
            LOGGER.error(f'Unable to delete sound file [{sound_file}] - {repr(ex)}')

        LOGGER.debug(f'Speak thread {cls._speak_thread_id} ended.')
        cls._speak_thread_id = None
        cls._locked = False
        return ret

    @classmethod
    def _play(cls, sound_file: str, speed: float = 1.0) -> int:
        '''Play the sound file'''
        check_file = pathlib.Path(sound_file)
        if not check_file.is_file():
            msg = f'Sorry, sound file {sound_file} does not exist.'
            LOGGER.warning(msg)
            return -1

        if cls._is_VLC_installed():        
            LOGGER.debug(f'Playing file: {cls._VLC} --intf dummy --rate {speed} --play-and-exit {sound_file}')
            if helper.is_windows():
                ret = os.system(f'"{cls._VLC}" --intf dummy --rate {speed} --play-and-exit {sound_file}')
            else:
                ret = os.system(f'{cls._VLC} --rate {speed} --play-and-exit {sound_file}')
        else:
            LOGGER.warning('Unable to play file, VLC not detected.')
            ret = -1

        return  ret

    @classmethod
    def _is_VLC_installed(cls) -> bool:

        if helper.is_windows():
            cls._VLC = cls.__chk_vlc_windows()
        else:
            cls._VLC = cls.__chk_vlc_linux()

        return cls._VLC is not None

    @classmethod
    def __chk_vlc_windows(cls) -> str:
        exe = 'vlc.exe'
        start_path = pathlib.Path(os.environ['ProgramFiles'])
        LOGGER.debug(f'- Searching for {exe} starting at {start_path}')
        target = helper.find_file(filenm=exe, search_path=start_path)
        if target is None:
            start_path = pathlib.Path(os.environ['ProgramFiles(x86)'])
            LOGGER.debug(f'- Searching for {exe} starting at {start_path}')
            target = helper.find_file(filenm=exe, search_path=start_path)
    
        return target
    
    @classmethod
    def __chk_vlc_linux(cls) -> str:
        start_path = pathlib.Path('/usr/bin')
        LOGGER.debug(f'- Searching for "cvlc" starting at {start_path}')
        return helper.find_file(filenm='cvlc', search_path=start_path)

    @classmethod
    def _is_file(cls, token: str) -> bool:
        check_file = pathlib.Path(token)
        try:
            is_file = check_file.is_file()
        except OSError:
            is_file = False
        return is_file
    
if __name__ == "__main__":
    import dt_tools.logger.logging_helper as lh
    from dt_tools.cli.demos.dt_misc_sound_demo import demo
    
    LOGGER.enable('dt_tools.misc')
    lh.configure_logger(log_level="DEBUG")
    demo()
